(*
 *
 * Copyright (c) 2006-2007, 
 *  Octavian Udrea	<udrea@umiacs.umd.edu>
 *  Cristian Lumezanu	<lume@cs.umd.edu>
 *  Jeffrey S. Foster	<jfoster@cs.umd.edu>
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. The names of the contributors may not be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *)

(* varunify.ml: Contains functions related to fact unification techniques - which is in fact a limited form of smbolic execution *)

open Types
open Rulefunc


(* This function prints a comparison operator *)
let printCompOp (op: comparisonOp) = 
    match op with
    EqOp -> print_string "="
    | LtOp -> print_string "<"
    | GtOp -> print_string ">"
    | LteOp -> print_string "<="
    | GteOp -> print_string ">="
    | NeqOp -> print_string "<>"
    ;;

(* This function prints an inVar *)
let printInVar (v: inVar) = 
    match v with
    InVar(x) -> Printf.printf "IN.%s" x.name
    | ArrayInVar(x,i) -> Printf.printf "IN.%s[%d]" x.name i
    ;;

(* This function prints an outVar *)
let printOutVar (v: outVar) = 
    match v with
    OutVar(x) -> Printf.printf "OUT.%s" x.name
    | ArrayOutVar(x,i) -> Printf.printf "OUT.%s[%d]" x.name i
    ;;
    
(* This function prints an outVar *)
let printRuleVar (v: ruleVar) = 
    match v with
    RuleVar(x) -> Printf.printf "?%s" x.name
    | ArrayRuleVar(x,i) -> Printf.printf "?%s[%d]" x.name i
    ;;

(* This function prints a var *)
let printVar (v: var) = 
    print_string v.name
    ;;
    
(* This function prints an allVar *)
let printAllVar (v: allVar) = 
    match v with
    AInVar(v) -> printInVar v
    | AOutVar(v) -> printOutVar v
    | ARuleVar(v) -> printRuleVar v
    | AVar(v) -> printVar v
    | AInt(v) -> print_int v
    ;;

(* This function prints an arithmetic operator *)
let printArithmeticOp (op: arithmeticOp) =
    match op with
    PlusOp -> print_string "+"
    | MinusOp -> print_string "-"
    | TimesOp -> print_string "*"
    | DivOp -> print_string "/"
    ;;

(* This function prints out a mixed expression *)
let rec printMixedExpr (e: mixedExpr) = 
    match e with
    MEInVar(v) -> printInVar v
    | MEOutVar(v) -> printOutVar v
    | MERuleVar(v) -> printRuleVar v
    | MEInt(v) -> print_int v
    | MEVar(v) -> printVar v
    | MixedExpr(op,e1,e2) -> 
	print_string "(";
	printArithmeticOp op;
	print_string " ";
	printMixedExpr e1;
	print_string " ";
	printMixedExpr e2;
	print_string ")"
    ;;

(* This function prints a fact to the output *)
let printFact (f: fact) = 
    match f with
    Nofact-> print_string("nofact")
    | Fact(op, v, e) -> print_string "(";
	printCompOp op;
	print_string " ";
	printAllVar v;
	print_string " ";
	printMixedExpr e;
	print_string ")"
    ;;

let rec _rec_printFactList (fl: factList) =
    match fl with
    [] -> print_string ""
    | h::r -> 
		print_string "["; 
		printFact h; 
		print_string "]"; 
		_rec_printFactList r
    ;;
    
(* This function prints a set of facts *)
let printFactList (flist: factList) = 
    print_string "{";
    _rec_printFactList flist;
    print_string "}"
    ;;

(* Transforms a inExpr to a mixedExpr *)
let rec in_to_mixedExpr (e: inExpr) =
    match e with 
    EInVar(v) -> MEInVar(v)
    | EInt(x) -> MEInt(x)
    | ERuleVar(v) -> MERuleVar(v)
    | InExpr(op,e1,e2) -> MixedExpr(op,(in_to_mixedExpr e1),(in_to_mixedExpr e2))
    ;;

(* Transforms a premise into a fact *)
let premise_to_fact (p: premise) = 
    match p with
    RVarPremiseStr(rv,v) -> Fact(EqOp, ARuleVar(rv), MEVar(v))
    | RVarPremiseInt (rv,x) -> Fact(EqOp, ARuleVar(rv), MEInt(x))
    | Premise(op,inv,ine) -> Fact(op, AInVar(inv), (in_to_mixedExpr ine))
    ;;
    
(* Transforms a conclusion into a fact *)
let conclusion_to_fact (c : conclusion) = 
    match c with
    Set(rv,me) -> Fact(EqOp, ARuleVar(rv), me)
    | NoOut -> Nofact
    | Success -> Nofact
    | Fail -> Nofact
    | Conclusion(op,ov,ine) -> Fact(op, AOutVar(ov),(in_to_mixedExpr ine))
    ;;

(* transforms a set of conclusions to a set of facts *)
let rec conclusionList_to_facts (c: conclusion list) : factList = 
    match c with
    [] -> []
    | h::r -> (conclusion_to_fact h)::(conclusionList_to_facts r)
    ;;

(* This function verifies if a var appears anywhere in an allVar *)
let appearsInAllVar (v: var) (av: allVar) = 
    match av with
    AInVar(x) -> (match x with
		  InVar(y) -> if y.name = v.name then true else false
		  | ArrayInVar(y,_) -> if y.name = v.name then true else false
		  )
    | AOutVar(x) -> (match x with
		    OutVar(y) -> if y.name = v.name then true else false
		    | ArrayOutVar(y,_) -> if y.name = v.name then true else false
		    )
    | ARuleVar(x) -> (match x with
		    RuleVar(y) -> if y.name = v.name then true else false
		    | ArrayRuleVar(y,_) -> if y.name = v.name then true else false
		     )
    | AInt(x) -> false
    | AVar(x) -> if x.name = v.name then true else false
    ;;

(* This function checks whether a variable appears anywhere in a mixed expression *)
let rec appearsInMixedExpr (v: var) (e: mixedExpr) = 
    match e with
    MEInVar(x) -> (match x with
		  InVar(y) -> if y.name = v.name then true else false
		  | ArrayInVar(y,_) -> if y.name = v.name then true else false
		  )
    | MEOutVar(x) -> (match x with
		    OutVar(y) -> if y.name = v.name then true else false
		    | ArrayOutVar(y,_) -> if y.name = v.name then true else false
		    )
    | MERuleVar(x) -> (match x with
		    RuleVar(y) -> if y.name = v.name then true else false
		    | ArrayRuleVar(y,_) -> if y.name = v.name then true else false
		     )
    | MEInt(x) -> false
    | MEVar(x) -> if x.name = v.name then true else false
    | MixedExpr(_,e1,e2) -> (appearsInMixedExpr v e1) || (appearsInMixedExpr v e2)
    ;;

(* This function takes a var and checks whether that var appears anywhere in a fact *)
let appearsInFact (v: var) (f: fact) =
    match f with
    Nofact -> false
    | Fact(_,x,e) -> (appearsInAllVar v x) || (appearsInMixedExpr v e)
    ;;

(* This function verifies if a fact f can be killed by newfact - meaning that the lhs of newfact (except for ints) appears somewhere in f *)
let canKill (f: fact) (newf: fact) = 
    match newf with
    Nofact -> false
    | Fact(_,x,_) ->
	    (match x with
	    AVar(v)-> appearsInFact v f
	    | AOutVar(v) -> (match v with
			    OutVar(y)-> appearsInFact y f
			    | ArrayOutVar(y,_) -> appearsInFact y f
			    )
	    | AInVar(v) -> (match v with
			    InVar(y)-> appearsInFact y f
			    | ArrayInVar(y,_) -> appearsInFact y f
			    )
	    | ARuleVar(v) -> (match v with
			    RuleVar(y)-> appearsInFact y f
			    | ArrayRuleVar(y,_) -> appearsInFact y f
			    )
	    | AInt(x) -> false
	)
    ;;

(* This function kills all facts in the list that should be killed by newfact and returns the new list *)
let rec kill_facts (facts: factList) (newf: fact) =
    match facts with
    [] -> []
    | h::r -> if canKill h newf then
		 kill_facts r newf
	      else
	        h::(kill_facts r newf)
    ;;

(* Checks for equality between two inVars *)
let equalsInVar (v1: inVar) (v2: inVar) =
    match v1 with
	InVar(x) -> (
			match v2 with
			InVar(y) -> if x.name = y.name then true else false
			| ArrayInVar(_,_) -> false
		    )
	| ArrayInVar(x,i) -> (match v2 with 
			InVar(y) -> false
			| ArrayInVar(y,j) -> if x.name = y.name then
						if i = j then true else false
					    else false
		    )
	;;

(* Checks for equality between two outVars *)
let equalsOutVar (v1: outVar) (v2: outVar) =
    match v1 with
	OutVar(x) -> (
			match v2 with
			OutVar(y) -> if x.name = y.name then true else false
			| ArrayOutVar(_,_) -> false
		    )
	| ArrayOutVar(x,i) -> (match v2 with 
			OutVar(y) -> false
			| ArrayOutVar(y,j) -> if x.name = y.name then
						if i = j then true else false
					    else false
		    )
	;;
	
(* Checks for equality between two ruleVars *)
let equalsRuleVar (v1: ruleVar) (v2: ruleVar) =
    match v1 with
	RuleVar(x) -> (
			match v2 with
			RuleVar(y) -> if x.name = y.name then true else false
			| ArrayRuleVar(_,_) -> false
		    )
	| ArrayRuleVar(x,i) -> (match v2 with 
			RuleVar(y) -> false
			| ArrayRuleVar(y,j) -> if x.name = y.name then
						if i = j then true else false
					    else false
		    )
	;;
	
(* Checks for equality between two allVars *)
let equalsAllVar (v1: allVar) (v2: allVar) =
    match v1 with
	AInVar(inv1) -> (match v2 with
			AInVar(inv2)-> equalsInVar inv1 inv2
			| AOutVar(_) -> false
			| ARuleVar(_) -> false
			| AInt(_) -> false
			| AVar(_) -> false
			)
	| AOutVar(inv1) -> (match v2 with
			AOutVar(inv2)-> equalsOutVar inv1 inv2
			| AInVar(_) -> false
			| ARuleVar(_) -> false
			| AInt(_) -> false
			| AVar(_) -> false
			)
	| ARuleVar(inv1) -> (match v2 with
			ARuleVar(inv2)-> equalsRuleVar inv1 inv2
			| AInVar(_) -> false
			| AOutVar(_) -> false
			| AInt(_) -> false
			| AVar(_) -> false
			)
	| AVar(inv1) -> (match v2 with
			AVar(inv2)-> if inv1.name = inv2.name then true else false
			| AInVar(_) -> false
			| AOutVar(_) -> false
			| AInt(_) -> false
			| ARuleVar(_) -> false
			)
	| AInt(inv1) -> (match v2 with
			AInt(inv2)-> if inv1 = inv2 then true else false
			| AInVar(_) -> false
			| AOutVar(_) -> false
			| AVar(_) -> false
			| ARuleVar(_) -> false
			)
	;;

(* This function finds a replacement for the given InVar in the fact list; It returns the mixed expression that replaces it *)
let rec findReplInVar (facts: factList) (v: inVar) = 
    match facts with
    [] -> MEInVar(v)
    | h::r -> 
	(match h with 
	Nofact-> findReplInVar r v
	| Fact(op,vv,e) -> if equalsAllVar (AInVar(v)) vv then
			    e
			  else findReplInVar r v
	)
    ;;

(* This function finds a replacement for the given OutVar in the fact list; It returns the mixed expression that replaces it *)
let rec findReplOutVar (facts: factList) (v: outVar) = 
    match facts with
    [] -> MEOutVar(v)
    | h::r -> 
	(match h with 
	Nofact-> findReplOutVar r v
	| Fact(op,vv,e) -> if equalsAllVar (AOutVar(v)) vv then
			    e
			  else findReplOutVar r v
	)
    ;;
    
(* This function finds a replacement for the given RuleVar in the fact list; It returns the mixed expression that replaces it *)
let rec findReplRuleVar (facts: factList) (v: ruleVar) = 
    match facts with
    [] -> MERuleVar(v)
    | h::r -> 
	(match h with 
	Nofact-> findReplRuleVar r v
	| Fact(op,vv,e) -> if equalsAllVar (ARuleVar(v)) vv then
			    e
			  else findReplRuleVar r v
	)
    ;;
    
(* This function finds a replacement for the given var in the fact list; It returns the mixed expression that replaces it *)
let rec findReplVar (facts: factList) (v: var) = 
    match facts with
    [] -> MEVar(v)
    | h::r -> 
	(match h with 
	Nofact-> findReplVar r v
	| Fact(op,vv,e) -> if equalsAllVar (AVar(v)) vv then
			    e
			  else findReplVar r v
	)
    ;;
    
(* This function searches the mixed expression and replace any atoms with any rhs of a fact containing the atom on the lhs - recursive after e *)
let rec replaceAtomsInMixedExpr (facts: factList) (e: mixedExpr) =
    match e with
    MEInVar(v) -> findReplInVar facts v
    | MEOutVar(v) -> findReplOutVar facts v
    | MERuleVar(v) -> findReplRuleVar facts v
    | MEVar(v) -> findReplVar facts v
    | MEInt(_) -> e
    | MixedExpr(op,e1,e2) -> MixedExpr(op, (replaceAtomsInMixedExpr facts e1), (replaceAtomsInMixedExpr facts e2))
    ;;

(* This function searches the fact f rhs for any variables x such that a fact x = something exists in the list, then it replaces x with something...; it returns the "replaced" newf *)
let replaceAtoms (facts: factList) (newf: fact) =
    match newf with
    Nofact -> Nofact
    | Fact(op, x, e) -> Fact(op, x, (replaceAtomsInMixedExpr facts e))
    ;;

(* Given an operator and two integers, it returns an integer that is the result of the operation *)
let doArithmetic (op: arithmeticOp) (x: int) (y: int) = 
    match op with
    PlusOp -> x + y
    | MinusOp -> x - y
    | TimesOp -> x * y
    | DivOp -> x / y
    ;;

(* This function is given a mixed expression, takes the rhs and tries to reduce any arithmetic expressions available (between ints) *)
let rec reduceArithmetic (e: mixedExpr) = 
    match e with
    MEInVar(_) -> e
    | MEOutVar(_) -> e
    | MERuleVar(_) -> e
    | MEInt(_) -> e
    | MEVar(_) -> e
    | MixedExpr(op,e1,e2) -> 
	    let re1 = reduceArithmetic e1 in
		let re2 = reduceArithmetic e2 in
		    (match re1 with
			MEInVar(_) -> (match re2 with
					MEInVar(_) -> MixedExpr(op,re1,re2)
					| MEOutVar(_) -> MixedExpr(op,re1,re2)
					| MERuleVar(_) -> MixedExpr(op,re1,re2)
					| MEVar(_) -> MixedExpr(op, re1, re2)
					| MixedExpr(_,_,_) -> MixedExpr(op, re1, re2)
					| MEInt(y) -> if (op = PlusOp) && (y = 0) then
							re1
						      else if (op = MinusOp) && (y = 0) then
						    		re1
							    else if (op = TimesOp) && (y = 0) then
								MEInt(0)
								 else if (op = TimesOp) && (y = 1) then
								    re1
								    else if (op = DivOp) && (y = 0) then
									failwith "Division by zero!!!"
									else MixedExpr(op, re1,re2)
				      )
			| MEOutVar(_) -> (match re2 with
					MEInVar(_) -> MixedExpr(op,re1,re2)
					| MEOutVar(_) -> MixedExpr(op,re1,re2)
					| MERuleVar(_) -> MixedExpr(op,re1,re2)
					| MEVar(_) -> MixedExpr(op, re1, re2)
					| MixedExpr(_,_,_) -> MixedExpr(op, re1, re2)
					| MEInt(y) -> if (op = PlusOp) && (y = 0) then
							re1
						      else if (op = MinusOp) && (y = 0) then
						    		re1
							    else if (op = TimesOp) && (y = 0) then
								MEInt(0)
								 else if (op = TimesOp) && (y = 1) then
								    re1
								    else if (op = DivOp) && (y = 0) then
									failwith "Division by zero!!!"
									else MixedExpr(op, re1,re2)
				      )
			| MERuleVar(_) -> (match re2 with
					MEInVar(_) -> MixedExpr(op,re1,re2)
					| MEOutVar(_) -> MixedExpr(op,re1,re2)
					| MERuleVar(_) -> MixedExpr(op,re1,re2)
					| MEVar(_) -> MixedExpr(op, re1, re2)
					| MixedExpr(_,_,_) -> MixedExpr(op, re1, re2)
					| MEInt(y) -> if (op = PlusOp) && (y = 0) then
							re1
						      else if (op = MinusOp) && (y = 0) then
						    		re1
							    else if (op = TimesOp) && (y = 0) then
								MEInt(0)
								 else if (op = TimesOp) && (y = 1) then
								    re1
								    else if (op = DivOp) && (y = 0) then
									failwith "Division by zero!!!"
									else MixedExpr(op, re1,re2)
				      )
			| MEVar(_) -> (match re2 with
					MEInVar(_) -> MixedExpr(op,re1,re2)
					| MEOutVar(_) -> MixedExpr(op,re1,re2)
					| MERuleVar(_) -> MixedExpr(op,re1,re2)
					| MEVar(_) -> MixedExpr(op, re1, re2)
					| MixedExpr(_,_,_) -> MixedExpr(op, re1, re2)
					| MEInt(y) -> if (op = PlusOp) && (y = 0) then
							re1
						      else if (op = MinusOp) && (y = 0) then
						    		re1
							    else if (op = TimesOp) && (y = 0) then
								MEInt(0)
								 else if (op = TimesOp) && (y = 1) then
								    re1
								    else if (op = DivOp) && (y = 0) then
									failwith "Division by zero!!!"
									else MixedExpr(op, re1,re2)
				      )
			| MixedExpr(_,_,_) -> (match re2 with
					MEInVar(_) -> MixedExpr(op,re1,re2)
					| MEOutVar(_) -> MixedExpr(op,re1,re2)
					| MERuleVar(_) -> MixedExpr(op,re1,re2)
					| MEVar(_) -> MixedExpr(op, re1, re2)
					| MixedExpr(_,_,_) -> MixedExpr(op, re1, re2)
					| MEInt(y) -> if (op = PlusOp) && (y = 0) then
							re1
						      else if (op = MinusOp) && (y = 0) then
						    		re1
							    else if (op = TimesOp) && (y = 0) then
								MEInt(0)
								 else if (op = TimesOp) && (y = 1) then
								    re1
								    else if (op = DivOp) && (y = 0) then
									failwith "Division by zero!!!"
									else MixedExpr(op, re1,re2)
				      )
			| MEInt(x) -> if (op = PlusOp) && (x = 0) then
					re2
				      else if (op = TimesOp) && (x = 0) then
				    	    MEInt(0)
					    else if (op = TimesOp) && (x = 1) then
						re2
						 else if (op = DivOp) && (x = 0) then
						    MEInt(0)
						    else
							(match re2 with
							MEInVar(_) -> MixedExpr(op,re1,re2)
						    	| MEOutVar(_) -> MixedExpr(op,re1,re2)
							| MERuleVar(_) -> MixedExpr(op,re1,re2)
							| MEVar(_) -> MixedExpr(op, re1, re2)
							| MixedExpr(_,_,_) -> MixedExpr(op, re1, re2)
							| MEInt(y) -> MEInt(doArithmetic op x y)
				    			 )
		    )
    ;;
	

(* This function reduces the arithmetic for the rhs of the fact *)
let reduceArithmetic_Fact (f: fact) = 
    match f with
    Nofact -> Nofact
    | Fact(op,v,e) -> Fact(op,v,(reduceArithmetic e))
    ;;

(* This function checks whether two mixed expressions are equal *)
let rec equalsMixedExpr (e1: mixedExpr) (e2: mixedExpr) = 
	match e1 with
	MEInVar(x) ->   (match e2 with
			    MEInVar(y) -> equalsInVar x y
			    | MEOutVar(_) -> false
			    | MERuleVar(_) -> false
			    | MEVar(_) -> false
			    | MEInt(_) -> false
			    | MixedExpr(_,_,_) -> false
			)
	| MEOutVar(x) -> (match e2 with
			    MEOutVar(y) -> equalsOutVar x y
			    | MEInVar(_) -> false
			    | MERuleVar(_) -> false
			    | MEVar(_) -> false
			    | MEInt(_) -> false
			    | MixedExpr(_,_,_) -> false
			 )
	| MERuleVar(x) ->(match e2 with
			    MERuleVar(y) -> equalsRuleVar x y
			    | MEOutVar(_) -> false
			    | MEInVar(_) -> false
			    | MEVar(_) -> false
			    | MEInt(_) -> false
			    | MixedExpr(_,_,_) -> false
			)
	| MEVar(x) -> (match e2 with
			    MEVar(y) -> if x.name = y.name then true else false
			    | MEOutVar(_) -> false
			    | MERuleVar(_) -> false
			    | MEInVar(_) -> false
			    | MEInt(_) -> false
			    | MixedExpr(_,_,_) -> false
			)
	| MEInt(x) -> (match e2 with
			    MEInt(y) -> if x = y then true else false
			    | MEOutVar(_) -> false
			    | MERuleVar(_) -> false
			    | MEVar(_) -> false
			    | MEInVar(_) -> false
			    | MixedExpr(_,_,_) -> false
			)
	| MixedExpr(op1,e11,e12) -> (match e2 with
			    MEInVar(_) -> false
			    | MEOutVar(_) -> false
			    | MERuleVar(_) -> false
			    | MEVar(_) -> false
			    | MEInt(_) -> false
			    | MixedExpr(op2,e21,e22) -> (op1 = op2) && (equalsMixedExpr e11 e21) && (equalsMixedExpr e12 e22)
			)
    ;;

(* This function checks whether two facts are identical *)
let equalsFacts (f1: fact) (f2: fact) = 
    match f1 with
    Nofact -> false
    | Fact(op1, v1, e1) -> (match f2 with
			Nofact -> false
			| Fact(op2,v2,e2) -> 
				(op1 = op2) && (equalsAllVar v1 v2) && (equalsMixedExpr e1 e2)
			)
    ;;

(* This function verifies whether a fact is in a fact list *)
let rec list_contains_fact (facts: factList) (f: fact) = 
    match facts with
    [] -> false
    | h::r -> if equalsFacts h f then true
	    else list_contains_fact r f
    ;;

(* Assumes facts are already unified and attempts to add a new fact to the set;
   It will unify newf with any pertinent facts in the list and produce a new fact
   list *)
(* ALGORITHM (short):
     1. Primary values are: in/out variables, rule variables and integers
     2. Check for any entities in the new fact that are not primary values
     3. Take the expression on the rhs of newfact
        3.1. For any atomic entity e in there, try and find a fact with = where e is on the lhs.
	3.2. Replace e with the rhs in the expression. NOTE: there is no need to repeat the process since we assume that we can no longer replace stuff on the rhs of existing facts (since they were unified earlier)
	3.3. Reduce any arithmetic expressions possible (by performing the operations).
*)
let fact_unify (facts: factList) (newf: fact) = 
    let newf2 = replaceAtoms facts newf in
	let newf3 = reduceArithmetic_Fact newf2 in
	    match newf3 with
	    Nofact -> facts
	    | Fact(_,_,_) -> if list_contains_fact facts newf3 then
				facts
			     else newf3::facts
	;;
    
(* This function adds a new fact to the fact list; first it kills all the facts that can be killed by the new fact, then it performs unification/reduction *)
let fact_add (facts: factList) (newf: fact) =
    fact_unify (kill_facts facts newf) newf
    ;;

(* Transforms a list of premises into a list of facts *)
let rec premiseList_to_fact (p: premise list) : (fact list) =
    match p with
    [] -> []
    | h::r -> fact_add (premiseList_to_fact r) (premise_to_fact h)
    ;;
    
(* Transforms a list of conclusions into a list of facts *)
let rec conclusionList_to_fact (c: conclusion list) : (fact list) = 
    match c with
    [] -> []
    | h::r -> fact_add (conclusionList_to_fact r) (conclusion_to_fact h)
    ;;

(* Concatenates two fact lists - NOT associative *)
let rec concatFactList (f1: fact list) (f2: fact list) = 
    match f2 with
    [] -> f1
    | h::r -> concatFactList (fact_add f1 h) r
    ;;

(* Intersects two fact lists *)
let rec intersectFactList (f1: fact list) (f2: fact list) = 
    match f1 with
    [] -> []
    | h::r -> if list_contains_fact f2 h then
		h::(intersectFactList r f2)
	      else intersectFactList r f2	 
    ;;

(* Checks whether f1 is a subset of f2 *)
let rec subsetFactList (f1: fact list) (f2: fact list) = 
    match f1 with
    [] -> true
    | h::r -> if list_contains_fact f2 h then
		    subsetFactList r f2
	      else false
    ;;

(* Checks whether two fact lists are identical *)
let equalsFactList (f1: fact list) (f2: fact list) = 
    (subsetFactList f1 f2) && (subsetFactList f2 f1)
    ;;