(*
 * Description:
 *
 * Copyright (c) 2000 by
 *  David Blei			blei@cs.berkeley.edu
 *  Chris Harrelson		chrishtr@cs.berkeley.edu
 *  Ranjit Jhala		jhala@cs.berkeley.edu
 *  Rupak Majumdar		rupak@cs.berkeley.edu
 *  George C. Necula	necula@cs.berkeley.edu
 *  Shree P. Rahul		sprahul@cs.berkeley.edu
 * 	Wes Weimer			weimer@cs.berkeley.edu
 *  Dror Weitz			dror@cs.berkeley.edu
 *   
 * All rights reserved.  Permission to use, copy, modify and distribute
 * this software for research purposes only is hereby granted, 
 * provided that the following conditions are met: 
 * 1.  Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 * 3. The name of the authors may not be used to endorse or promote products derived from 
 * this software without specific prior written permission. 
 *
 * DISCLAIMER:
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR IMPLIED 
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *)
module E = Errormsg
module I = Input
module D = Dag
module S = Satproc
module P = Proof


open Pretty

let debug = ref false 

let engineIdent = "EG"                  (* This is the identifier of the 
                                           * engine itself *)
    
(* The engine will make assertions and each assertion 
 * will be assigned a unique id. The proofs of these 
 * assertions are kept in a hash table. We keep an undo 
 * stack where each element specifies *)

                                        (* The engine maintains an undoStack *)
type undoElem = 
    Snapshot                            (* Marks a snapshot*)
  | Assume of int * P.proof             (* Marks an assumption along with the 
                                         * message id and its proof *)

let undoStack : undoElem list ref = ref []


exception Failure

                                        (* Initialize the engine with the 
                                         * satisfiability procedures *)
let prove (satprocs : S.entries list) : 
          (string -> Input.pred -> Proof.proof)  = 

  ignore (printf "Using satprocs: %a\n"
            (docList (chr ',') (fun sp -> text sp.S.ident)) satprocs);

  let msgid = ref 0 in                  (* Initialize a message id counter *)
                                        (* The global snapshot function *)
  let snapshot () = 
    List.iter (fun sp -> sp.S.snapshot ()) satprocs;
		Subgoal.snapshot () ; (* [weimer] subgoal manager snapshot *)
    undoStack := Snapshot :: !undoStack
  in
                                        (* The global undo function *)
  let undo () = 
    List.iter (fun sp -> sp.S.undo ()) satprocs;
                                        (* Now pop the undoStack *)
		Subgoal.undo () ; (* [weimer] subgoal manager support *)
    let rec loop = function
        Snapshot :: rest -> undoStack := rest
      | Assume(i,p) :: rest -> 
					(if (!debug) then
					ignore (Pretty.printf " (%d,%a)" i Proof.dp p)) ;
					loop rest
      | [] -> E.s (E.bug "Engine tries to pop an empty stack")
    in
		(if (!debug) then ignore (Pretty.printf "EG: undoing")) ; 
    loop !undoStack ;
		(if (!debug) then ignore (Pretty.printf ".\n"))
  in

                                        (* This is the proof generator *)
  let proofGen (msg : S.message) : P.proof = 
    let ownerid = msg.S.owner in
                                        (* Check first if an Engine message *)
    if ownerid = engineIdent then
      let msgid = msg.S.id in
      let rec loop = function
          Assume (id,prf) :: _ when id = msgid -> prf
        | _ :: rest -> loop rest
        | [] -> 
								() ;
								E.s (E.bug "Engine cannot find proof for message %a" 
                       S.dm msg)
      in
      loop !undoStack
    else 
      let owner = 
        try List.find (fun sp -> sp.S.ident = ownerid) satprocs 
        with Not_found -> E.s (E.bug "cannot find satproc %s" ownerid)
      in
      owner.S.proveIt msg
  in
  S.proveIt := proofGen;  (* Publicize the proof generator *)

                                        (* The global assume function. Might 
                                         * raise S.Contradiction *)
  let assume (atom : I.atom) (isgoal : bool) (prf : P.proof) = 
                                        (* First must internalize the 
                                         * predicate *)
    let atommsgbody = match atom with 
      I.Eq(e1,e2) -> 
        let e1' = D.internExp e1 in
        let e2' = D.internExp e2 in
        S.Equality (e1', e2')
    | e -> S.Literal (D.internAtom e)
    in
                                        (* Now create a message *)
    let atommsg = 

      { S.body    = atommsgbody;
        S.owner   = engineIdent;        (* Owned by the engine *)
        S.id      = !msgid;             (* Engine does not use the id's *)
        S.goal    = isgoal;
      } 
    in
		(* [weimer @ 3-22-00] since the engine is asserting this 
		 * message, it must keep around a proof for it on the 
		 * undoStack.  *)
    undoStack := (Assume(!msgid, prf)) :: !undoStack ;
		(if (!debug) then 
		ignore (Pretty.printf "EG: have proof id %d, hyp = %a for %a\n" !msgid
			Proof.dp prf Input.da atom )) ;
		msgid := !msgid + 1 ;
		(* [weimer @ 3-22-00] ends *)
    let rec fixpoint msgs : unit =
                                        (* Send the messages to all 
                                         * satprocs. Collect everything they 
                                         * return *)
      let acc =
	List.fold_left
	  (fun acc sp -> List.fold_left sp.S.assume acc msgs)
	  []
	  satprocs
      in
      match acc with			(* Do this until no more messages *)
	[] -> ()
      | _ -> fixpoint (List.rev acc)
    in
    fixpoint [atommsg]
  in
    
                                        (* And this is the inversion function 
                                         * *)
  let rec inversion = function
      I.True -> P.G("truei", [])
    | I.False -> raise Failure
    | I.And(p1,p2) -> P.G("andi", [ P.P (inversion p1);
                                    P.P (inversion p2)])
    | I.All(v,p) -> P.Alli(v, inversion p)
    | I.Imp(p1,p2) -> begin
        snapshot ();
        try
          (* Create a new assumption variable *)
          let assumprf = I.newVar "H" in
          let prf = 
            P.Impi(assumprf, p1,
                   try
                     hypothesis p1 (P.Hyp assumprf);(* Process the hypothesis*)
                     (* no contradiction in the assumptions *)
                     inversion p2
                   with S.Contradiction pfalse -> 
                     P.G("falsee", [P.F p2; P.P pfalse]))
          in
          undo ();
          prf
        with e -> begin
          undo ();
          raise e
        end
    end
    (* Now try to prove an atom *)
    | I.Atom atom -> begin
        snapshot ();
        try
          let assumvar = I.newVar "C" in
          let prf = 
            P.Contra(assumvar, atom,
                     try
                       assume atom true (P.Hyp assumvar);
											 Equality.expensiveChecks () ; (* try CC once more *)
											 tryAllSubgoals () ; (* [weimer] subgoal support *) 
                       raise Failure
                     with S.Contradiction pfalse -> pfalse)
          in
          undo ();
          prf
        with e -> 
          undo ();
          raise e
    end
    | _ -> E.s (E.unimp "engine.inversion")

		(* [weimer] subgoal support *)
		and tryAllSubgoals () = 
			while not (Subgoal.empty ()) do 
				match (Subgoal.getNext ()) with
					Some(sg) -> 
				( try
					raise (S.Contradiction( sg.Subgoal.proof_gen 
							(List.map inversion sg.Subgoal.goal) ));
					with
					Failure -> ignore () 
				) 
			|       _ -> E.s (E.bug "engine.subgoal.getNext")
			done 



  and hypothesis hyp prf = 
    match hyp with (* Now check the proof *)
      I.True -> ()
    | I.False -> raise (S.Contradiction prf)
    | I.And(p1, p2) -> 
        hypothesis p1 (P.G("andel", [P.P prf]));
        hypothesis p2 (P.G("ander", [P.P prf]))
    | I.Atom atom -> assume atom false prf 
    | _ -> E.s (E.unimp "engine.hypothesis")
  in

  (* This is the real prover *)
  fun name pred -> 
    ignore (printf "Proving goal %s:\n %a\n" name I.dp pred);
    let prf = Printexc.catch inversion pred in
    let olf = !I.printLFSyntax in
    I.printLFSyntax := true;
    ignore (fprintf !E.proofChannel
              "pf%s : pf @[%a@] \n%% The proof@! :: @[%a@].\n" 
              name I.dp pred P.dp prf);
    I.printLFSyntax := olf;

    let ipred = P.infer [] [] prf in
    if not (I.samep pred ipred) then
      E.s (E.error "Inferred predicate does not match the predicate to prove")
    else ignore (printf "\nProof checked successfully.\n");


    prf
  

      

