(*
 *
 * Copyright (c) 2006-2007, 
 *  Octavian Udrea	<udrea@umiacs.umd.edu>
 *  Cristian Lumezanu	<lume@cs.umd.edu>
 *  Jeffrey S. Foster	<jfoster@cs.umd.edu>
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. The names of the contributors may not be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *)

(* rmc.ml : The very core of it all: it contains the functions that do the actual rule checking *)

open Pretty
open Cil
open Eunify
open Rmcfind
open Rmcutils
open Rulefunc
open Types

(* This function obtains the initial set of facts for running any dataflow, which comes from global initializers in a Cil.file *)
let getInitialFacts (f: Cil.file) = 
    getFactFromGlobalList f.globals
    ;;

(* This function locates the statement facts for a specified ID and returns the list of facts *)
let rec findStmtFacts (sfacts: stmtFact list) (sid: int) : factList = 
    match sfacts with
    [] -> []
    | h::r -> 
	begin
	    match h with (x,fl) -> if x = sid then fl else findStmtFacts r sid
	end
    ;;
    
(* This function removes the facts for a given statement from the list and returns the new list *)
let rec removeStmtFacts (sfacts: stmtFact list) (sid: int) : (stmtFact list) = 
    match sfacts with
    [] -> []
    | h::r -> 
	begin
	    match h with (x,fl) -> if x = sid then (removeStmtFacts r sid) else h::(removeStmtFacts r sid)
	end
    ;;

(* This function takes a list of stmtFacts, a statement ID and a list of facts. It must add the list of facts to the facts for the specified statement *)
let addFactsOneStmt (sfacts: stmtFact list) (sid: int) (fl: factList) : (stmtFact list) = 
    let crFacts = findStmtFacts sfacts sid in
	let newStmtFactList = removeStmtFacts sfacts sid in
	    let newFactPair = (sid,(concatFactList crFacts fl)) in
	       let resFactList = newFactPair::newStmtFactList in
	    	    resFactList
    ;;

(* This function does the same thing as above, but for a list of statements *)
let rec addFactsMultipleStmts (sfacts: stmtFact list) (stmts: Cil.stmt list) (fl: factList) : (stmtFact list) = 
    match stmts with
    [] -> sfacts
    | h::r -> addFactsMultipleStmts (addFactsOneStmt sfacts h.sid fl) r fl
    ;;

(* This function takes a context and it must add a new set of facts to all startpoints *)
let addFactsStmts (fl: factList) (stmts: Cil.stmt list) (c: ruleContext) : ruleContext = 
    { startpoints = c.startpoints;
      relVars = c.relVars;
      facts = addFactsMultipleStmts c.facts stmts fl;
      endpoints = c.endpoints;
    }
    ;;

(* Takes a list of patterns and returns a list containing only the input patterns *)
let rec getInputPatterns (p: pattern list) : (pattern list) =
    match p with
    [] -> []
    | h::r -> begin
		match h with InPat(_,_) -> h::(getInputPatterns r)
			    | _ -> getInputPatterns r
	      end
    ;;

(* Takes a list of patterns and returns a list containing only the success patterns *)
let rec getSuccPatterns (p: pattern list) : (pattern list) =
    match p with
    [] -> []
    | h::r -> begin
		match h with SuccPatF(_,_) -> h::(getSuccPatterns r)
			    | SuccPatR(_) -> h::(getSuccPatterns r)
			    | _ -> getSuccPatterns r
	      end
    ;;
    
(* Takes a list of patterns and returns a list containing only the fail patterns *)
let rec getFailPatterns (p: pattern list) : (pattern list) =
    match p with
    [] -> []
    | h::r -> begin
		match h with FailPatF(_,_) -> h::(getFailPatterns r)
			    | FailPatR(_) -> h::(getFailPatterns r)
			    | _ -> getFailPatterns r
	      end
    ;;
    
(* Takes a list of patterns and returns a list containing only the output patterns *)
let rec getOutputPatterns (p: pattern list) : (pattern list) =
    match p with
    [] -> []
    | h::r -> begin
		match h with OutPat(_,_) -> h::(getOutputPatterns r)
			    | _ -> getOutputPatterns r
	      end
    ;;
    
(* Creates a list with the successors of all the given statements *)
let rec getSuccessorsStmtList (sl: Cil.stmt list) : (Cil.stmt list) = 
    match sl with
    [] -> []
    | h::r -> concatStmtList h.succs (getSuccessorsStmtList r)
    ;;

(* Function takes an initial set of facts, an instruction list and determines the list of facts at the end of the list *)
let rec _recDataFlowInstructionList (il: Cil.instr list) (f: factList) (innames: string list) (outnames: string list) : factList = 
    match il with
    [] -> f
    | h::r -> let newf = getFact h innames outnames in
		_recDataFlowInstructionList r (fact_add f newf) innames outnames	
    ;;

(* Gets a ruleContext and sets the new fact list for statement with id sid and returns the new context *)
let changeContextStmtFacts (c: ruleContext) (sid: int) (fl: factList) : ruleContext = 
    let exfacts = findStmtFacts c.facts sid in
    let frem = removeStmtFacts c.facts sid in
	let newFactPair = (sid,(concatFactList exfacts fl)) in
	    let resf = newFactPair::frem in
		{ startpoints = c.startpoints;
		  relVars = c.relVars;
		  facts = resf;
		  endpoints = c.endpoints;
		}
    ;;

(* Function intersects all existing facts for a list of statements, but only for those that are in list L *)
let rec intersectPred (c: ruleContext) (slist: Cil.stmt list) (ll: Cil.stmt list): factList = 
    match slist with
    [] -> []
    | h::r -> if isStmtInList h ll then
		let fl = findStmtFacts c.facts h.sid in
		    intersectFactList fl (intersectPred c r ll)
	      else intersectPred c r ll
    ;;

(* Converts a comparison operator *)
let toComparisonOp (op: Cil.binop) : comparisonOp =
    match op with
    Lt -> LtOp
    | Gt -> GtOp
    | Le -> LteOp
    | Ge -> GteOp
    | Eq -> EqOp
    | Ne -> NeqOp
    | _ -> failwith "Wrong operator in toComparisonOp...\n"
    ;;

(* Converts a comparison operator *)
let toInvComparisonOp (op: Cil.binop) : comparisonOp =
    match op with
    Lt -> GteOp
    | Gt -> LteOp
    | Le -> GtOp
    | Ge -> LtOp
    | Eq -> EqOp
    | Ne -> NeqOp
    | _ -> failwith "Wrong operator in toInvComparisonOp...\n"
    ;;
    
(* Try to see if an expression can be changed to a fact *)
let checkExpToFact (e: Cil.exp) (innames: string list) (outnames: string list) = 
    match e with
    BinOp(op,e1,e2,_) ->
	let opOk = 
	(
	    match op with
	    Lt -> true
	    | Gt -> true
	    | Le -> true
	    | Ge -> true
	    | Eq -> true
	    | Ne -> true
	    | _ -> false
	) in
	    if opOk then
		match e1 with
		 Lval(lv) -> if checkLval lv innames outnames then
				if checkExpToMixedExpr e1 innames outnames then
				    let lva = lvalToAllVar lv innames outnames in
					let emix = expToMixedExpr e1 innames outnames in
					    true					
				else false
			     else false
		 | _ -> if checkExpToMixedExpr e1 innames outnames then
			    (
				match e2 with 
				Lval(lv) ->  if checkLval lv innames outnames then
						true
					     else false
				| _ -> false
			    )
			else false
	    else false
    | _ -> false
    ;;

(* Convert an expression to a fact *)
let expToFact (e: Cil.exp) (innames: string list) (outnames: string list) : fact= 
    match e with
    BinOp(op,e1,e2,_) ->
	let opOk = 
	(
	    match op with
	    Lt -> true
	    | Gt -> true
	    | Le -> true
	    | Ge -> true
	    | Eq -> true
	    | Ne -> true
	    | _ -> false
	) in
	    if opOk then
		match e1 with
		 Lval(lv) -> if checkLval lv innames outnames then
				if checkExpToMixedExpr e1 innames outnames then
				    let lva = lvalToAllVar lv innames outnames in
					let emix = expToMixedExpr e1 innames outnames in
					    Fact((toComparisonOp op),lva,emix)					
				else Nofact
			     else Nofact
		 | _ -> if checkExpToMixedExpr e1 innames outnames then
			    (
				match e2 with 
				Lval(lv) ->  if checkLval lv innames outnames then
    				    		let lva = lvalToAllVar lv innames outnames in
						    let emix = expToMixedExpr e2 innames outnames in
							Fact((toInvComparisonOp op),lva,emix)					
					     else Nofact
				| _ -> Nofact
			    )
			else Nofact
	    else Nofact
    | _ -> Nofact
    ;;

(* Gets the last element from a list *) 
let rec getLast l = 
    match l with
    [] -> failwith "No elements in getLast...\n"
    | h::r -> if (list_count r) = 0 then
		h
	      else getLast r
    ;;

(* find endpoints in an endpoint list *)
let rec findEndpointsInList (endp: ruleEndp list) (rname: string) : (Cil.stmt list) = 
    match endp with
    [] -> []
    | h::r -> (match h with (n,l) -> if rname = n then l else findEndpointsInList r rname
	      )
    ;;    

(* find the list of endpoints for a given rule in the current context *)
let rec findEndpoints (c: ruleContext) (rname: string) : (Cil.stmt list) = 
    findEndpointsInList c.endpoints rname
    ;;

(* remove the endpoints for a given rule *)
let rec removeEndpointsInList (endp: ruleEndp list) (rname: string) : (ruleEndp list) = 
    match endp with 
    [] -> []
    | h::r -> (match h with (n,l)->
		 if n = rname then
		    removeEndpointsInList r rname
		 else h::(removeEndpointsInList r rname)
	      )
    ;;
    
(* remove endpoints for a given rule from the context *)
let removeEndpoints (c: ruleContext) (rname: string) : ruleContext = 
    let newE = removeEndpointsInList c.endpoints rname in
	{ startpoints = c.startpoints;
	  relVars = c.relVars;
	  facts = c.facts;
	  endpoints = newE
	}
    ;;
    
(* Change the context by adding an endpoint to the current rule *)
let changeContextAddEndp (c: ruleContext) (endp: Cil.stmt) (rname: string) : ruleContext = 
    let newE = findEndpoints c rname in
	let newC = removeEndpoints c rname in
	    let newEPair = (rname,endp::newE) in
		{
		    startpoints = newC.startpoints;
		    relVars = newC.relVars;
		    facts = newC.facts;
		    endpoints = newEPair::(newC.endpoints)
		}
    ;;

(* Checks to see whether the endpoints for a certain rule have changed *)
let hasEndpointChanged (c1: ruleContext) (c2: ruleContext) (rname:string) = 
(*    let e1 = findEndpoints c1 rname in
	let e2 = findEndpoints c2 rname in
	    if (list_count e1) != (list_count e2) then true
	    else false *)
	    false
    ;;

(* Do dataflow on one statement only *)
let rec _recDataFlowOneStatement (s: Cil.stmt) (ll: Cil.stmt list) (c: ruleContext) (endp: Cil.stmt list) (innames: string list) (outnames: string list) (rname: string) (ruleend: Cil.stmt list) : ruleContext = 
    match s.skind with
	Instr(il) -> let newC = changeContextStmtFacts c s.sid (concatFactList (intersectPred c s.preds ll) (_recDataFlowInstructionList il (findStmtFacts c.facts s.sid) innames outnames)) in
			    if isStmtInList s ruleend then
			      changeContextAddEndp newC s rname
			    else newC
	| Return(_,_) -> let newC = changeContextStmtFacts c s.sid (intersectPred c s.preds ll) in
			    if isStmtInList s ruleend then
			      changeContextAddEndp newC s rname
			    else newC
	| Goto(_,_) -> let newC = changeContextStmtFacts c s.sid (intersectPred c s.preds ll) in
			    if isStmtInList s ruleend then
			      changeContextAddEndp newC s rname
			    else newC
	| Break(_) -> let newC = changeContextStmtFacts c s.sid (intersectPred c s.preds ll) in
			    if isStmtInList s ruleend then
			      changeContextAddEndp newC s rname
			    else newC
	| Continue(_) -> let newC = changeContextStmtFacts c s.sid (intersectPred c s.preds ll) in
			    if isStmtInList s ruleend then
			      changeContextAddEndp newC s rname
			    else newC
	| If(e,bt,bf,_) -> 
	
	    let pred = intersectPred c s.preds ll in
    		if checkExpToFact e innames outnames then
		    let fexp = expToFact e innames outnames in
			if list_contains_fact pred fexp then
			    (* only the true branch *)
			    let newC = _recDataFlowMultipleStmts bt.bstmts (concatStmtList ll (s::[])) c endp innames outnames rname ruleend in
    				    if (list_count bt.bstmts) > 0 then
					let lstmt = getLast bt.bstmts in
					    changeContextStmtFacts newC s.sid (findStmtFacts newC.facts lstmt.sid)
				    else newC			
		       else
		       (* both branches and intersect at the end *)
		       let newC = _recDataFlowMultipleStmts bt.bstmts (concatStmtList ll (s::[])) c endp innames outnames rname ruleend in
			    (* both branches *)
		    	      let newC2 = _recDataFlowMultipleStmts bf.bstmts (concatStmtList ll (s::[])) newC endp innames outnames rname ruleend in
				    (* both branches *)
				    if (list_count bt.bstmts) > 0 then
					let tlast = getLast bt.bstmts in
					    if (list_count bf.bstmts) > 0 then
						let flast = getLast bf.bstmts in
						    changeContextStmtFacts newC2 s.sid (intersectFactList (findStmtFacts newC2.facts tlast.sid) (findStmtFacts newC2.facts flast.sid))
					    else
					    (* no false statements *)
					    changeContextStmtFacts newC2 s.sid (findStmtFacts newC2.facts tlast.sid)
				    else
				    (* no true statement, still have to check whether false is there *)
					if (list_count bf.bstmts) > 0 then
					    let flast = getLast bf.bstmts in
						changeContextStmtFacts newC2 s.sid (findStmtFacts newC2.facts flast.sid)
					else
					(* do nothing to C2*)
					newC2
		else
		       (* both branches and intersect at the end *)
		       let newC = _recDataFlowMultipleStmts bt.bstmts (concatStmtList ll (s::[])) c endp innames outnames rname ruleend in
			    (* both branches *)
		    	      let newC2 = _recDataFlowMultipleStmts bf.bstmts (concatStmtList ll (s::[])) newC endp innames outnames rname ruleend in
				    (* both branches *)
				    if (list_count bt.bstmts) > 0 then
					let tlast = getLast bt.bstmts in
					    if (list_count bf.bstmts) > 0 then
						let flast = getLast bf.bstmts in
						    changeContextStmtFacts newC2 s.sid (intersectFactList (findStmtFacts newC2.facts tlast.sid) (findStmtFacts newC2.facts flast.sid))
					    else
					    (* no false statements *)
					    changeContextStmtFacts newC2 s.sid (findStmtFacts newC2.facts tlast.sid)
				    else
				    (* no true statement, still have to check whether false is there *)
					if (list_count bf.bstmts) > 0 then
					    let flast = getLast bf.bstmts in
						changeContextStmtFacts newC2 s.sid (findStmtFacts newC2.facts flast.sid)
					else
					(* do nothing to C2*)
					newC2
				    
	| Switch(e,b,sl,_) -> c
	| Loop(b,_,_,_) -> let newC = _recDataFlowMultipleStmts b.bstmts (concatStmtList ll (s::[])) c endp innames outnames rname ruleend in
				if (list_count b.bstmts) > 0 then
			    	    let lstmt = getLast b.bstmts in
			    		changeContextStmtFacts newC s.sid (findStmtFacts newC.facts lstmt.sid)
		    		else newC
	| Block(b) -> let newC = _recDataFlowMultipleStmts b.bstmts (concatStmtList ll (s::[])) c endp innames outnames rname ruleend in
				if (list_count b.bstmts) > 0 then
				    let lstmt = getLast b.bstmts in
					changeContextStmtFacts newC s.sid (findStmtFacts newC.facts lstmt.sid)
		    		else newC
	| _ -> c
and 
_recDataFlowMultipleStmts (slist: Cil.stmt list) (ll: Cil.stmt list) (c: ruleContext) (endp: Cil.stmt list) (innames: string list) (outnames: string list) (rname: string) (ruleend: Cil.stmt list) : ruleContext = 
    match slist with
    [] -> c
    | h::r -> let newC = _recDataFlowOneStatement h ll c endp innames outnames rname ruleend in
		 _recDataFlowMultipleStmts r (concatStmtList ll (h::[])) newC endp innames outnames rname ruleend
    ;;

(* Now we have to start moving through statements *)
let rec _recDataFlow (ll: Cil.stmt list) (wq: Cil.stmt list) (c: ruleContext) (endp: Cil.stmt list) (innames: string list) (outnames: string list) (rname: string) (ruleend: Cil.stmt list) : ruleContext = 
    match wq with
    [] -> c
    | h::r -> 
	let flist = findStmtFacts c.facts h.sid in
	    let newC = _recDataFlowOneStatement h ll c endp innames outnames rname ruleend in
		let newFlist = findStmtFacts newC.facts h.sid in
		    if equalsFactList flist newFlist then
			_recDataFlow (concatStmtList ll (h::[])) r newC endp innames outnames rname ruleend
		    else
			_recDataFlow (concatStmtList ll (h::[])) (concatStmtList r h.succs) newC endp innames outnames rname ruleend
			    
    ;;



(* This function performs the data flow analysis for a certain rule.
   It does the following:
   1. Detects the endpoints based on the current startpoints
   2. Runs dataflow between the startpoints and the endpoints and adds facts to the context (if needed); it also sets the endpoints for the curent rule so we know later for which
      statements to check the conclusions
   3. Detects the next startpoints (if any)
   4. Runs dataflow between the current endpoints and the next startpoints in order to prepare the context for checking the next rule
   NOTE to self: the dataflow may end either and the endpoints or at end of the CFG
*)
let makeDataFlow (c: ruleContext) (r: rule) (patterns: pattern list) : ruleContext = 
    let endpoints = rmcFind_StmtList c.startpoints patterns in (* current endpoints for this rule *)
	(* compute the next startpoints *)
	let inpats = getInputPatterns patterns in
	let outpats = getOutputPatterns patterns in
	let innames = 
	    (
		match inpats with
		[] -> []
		| h::r -> getPatternNames c.startpoints h
	    ) in
	let outnames = 
	    (
		match outpats with
		[] -> []
		| h::r -> getPatternNames c.startpoints h
	    ) in
	let newStart = rmcFind_StmtList endpoints (getInputPatterns patterns) in
    	    (* now we have to add the premises of r to the set of facts already in the context for all startpoints *)
	    let cBegin = addFactsStmts (premiseList_to_fact r.premises) c.startpoints c in (* created a new context as mentioned above *)
		(* now we have to move through the statements and do the dataflow - this should be a pain :) *)	
		let cEnd = _recDataFlow c.startpoints (getSuccessorsStmtList c.startpoints) cBegin newStart innames outnames r.rname.name endpoints in
		    { startpoints = newStart;
		      relVars = cEnd.relVars;
		      facts = cEnd.facts;
		      endpoints = cEnd.endpoints;
		    }
    ;;


(* This function verifies that a certain fact holds for a statement. It does this by unifying the fact with the existing facts, then 
   checking whether the set of facts changed *)
let verifyFactStmt (c: ruleContext) (f: fact) (s: Cil.stmt) = 
    let flist = findStmtFacts c.facts s.sid in
	let fadd = fact_add flist f in
	   equalsFactList flist fadd
    ;;

(* This function verifies that a certain fact holds for all statements in a statement list *)
let rec verifyFactStmtList (c: ruleContext) (f: fact) (slist: Cil.stmt list) = 
    match slist with
    [] -> true
    | h::r-> if verifyFactStmt c f h then verifyFactStmtList c f r else false
    ;;

(* This function verifies one conclusion of a given rule *)
let verifyConclusion (c: ruleContext) (rname: string) (con: conclusion) (pat: pattern list) =
    match con with
	Set(_,_) -> true
	| NoOut -> let res = allStmtInPatternList (findEndpoints c rname) (getInputPatterns pat) in
			Printf.printf "Verifying conclusion NO_OUTPUT for rule %s...\n" rname;
			res
	| Success -> let res = allStmtInPatternList (findEndpoints c rname) (getSuccPatterns pat) in
			Printf.printf "Verifying conclusion SUCCESS for rule %s...\n" rname;
			res
	| Fail -> let res = allStmtInPatternList (findEndpoints c rname) (getFailPatterns pat) in
			Printf.printf "Verifying conclusion FAILURE for rule %s...\n" rname;
			res
	| Conclusion(_,_,_) -> 	let f = conclusion_to_fact con in
				    Printf.printf "Verifying conclusion ";
				    printFact f;
				    Printf.printf " for rule %s...\n" rname;
				    let endp = findEndpoints c rname in
					let res = verifyFactStmtList c f endp in
					    res
	;;

(* This function verifies all conclusions in a list *)
let rec verifyConclusionList (c: ruleContext) (rname: string) (con: conclusion list) (pat: pattern list) = 
    match con with
    [] -> true
    | h::r -> if verifyConclusion c rname h pat then
		verifyConclusionList c rname r pat 
	      else false
    ;;

(* This function verifies all conclusions for a given rule *)
let verifyAllConclusions (c: ruleContext) (r: rule) (pat: pattern list) = 
    verifyConclusionList c r.rname.name r.conclusions pat
    ;;

(* This function extracts the side_effects from a rule's conclusion list *)
let rec extractSideEffects (c: conclusion list) : (conclusion list) = 
    match c with
    [] -> [] 
    | h::r -> (match h with Set(_,_) -> h::(extractSideEffects r)
			    | _ -> extractSideEffects r
	     )
    ;;	    
    
(* This function process all side effects of a rule. *)
let processSideEffects (c: ruleContext) (r: rule) : ruleContext = 
    let clist = extractSideEffects r.conclusions in
	let flist = conclusionList_to_facts clist in
	    addFactsStmts flist c.startpoints c
    ;;
    
(* This function verifies a rule - that means does the dataflow, does the conclusion verification and the side-effects *)
let verifyRule (c: ruleContext) (r: rule) (pat: pattern list) : ruleContext = 
    Printf.printf "Verifying tule %s begins...\n" r.rname.name;
    let newC = makeDataFlow c r pat in
	let res = verifyAllConclusions newC r pat in
	    let newC2 = processSideEffects newC r in
		newC2
    ;;

(* gets the first statement as a list *)
let getFirstStmt (f: Cil.fundec) : (Cil.stmt list) = 
    match f.sbody.bstmts with
    [] -> []
    | h::r -> h::[]
    ;;

(* Creates an initial context for when the program is run *)
let makeInitialContext (f: Cil.file) (pat: pattern list) : ruleContext = 
    let flist = getInitialFacts f in
	let fdec = locateMain f in
	    let bogus = 
		{ rname = {name = "bogus" };
		  premises = [];
		  conclusions = [];
		} in
		let pin = getInputPatterns pat in
		    let c = { startpoints = getFirstStmt fdec;
			      relVars = [];
			      facts = [];
			      endpoints = [];
			    } in
			    let newC = addFactsStmts flist c.startpoints c in
				let newC2 = makeDataFlow newC bogus pin in
				    (* we have to clear it now *)
				    let newC3 = { startpoints = newC2.startpoints;
						  relVars = [];
						  facts = newC2.facts;
						  endpoints = []
						} in newC3
    ;;

(* Special concatenation function: make sure the elements of l2 appear before those of l1 *)
let rec myspecial l1 l2 = 
    match l2 with
    [] -> l1
    | h::r -> myspecial (h::l1) r
    ;;

(* Given a rule list and a string list, it returns the rule list *)
let rec buildRuleList (rules: rule list) (names: string list) : (rule list) = 
    match names with
    [] -> []
    | h::r -> (getRuleByName rules h)::(buildRuleList rules r)
    ;;

(* This is the model checking recursive function *)
let rec _recModelCheck (rulep: program) (r: rule list) (c: ruleContext) (cfg: ruleCFG) (verified: rule list): ruleContext = 
    match r with
    [] -> c
    | h::rest -> 
	(
	let pred = buildRuleList rulep.rules (getCFGPredecessors cfg h.rname.name) in
	    match pred with
	    [] -> _recModelCheck rulep rest (verifyRule c h rulep.patterns) cfg (h::verified)
	    | t::z -> _recModelCheck rulep (myspecial r pred) c cfg verified
	)
    ;;

(* The BIG function - takes a program, a rule name and the CIL.file *)
let makeModelCheck (rulep: program) (rname: string) (f: Cil.file) = 
    let ctx = makeInitialContext f rulep.patterns in
	let cfg = createRuleCFG rulep in
	    let rcheck = getRuleByName rulep.rules rname in
		_recModelCheck rulep (rcheck::[]) ctx cfg []	
    ;;