(*
 *
 * Copyright (c) 2006-2007, 
 *  Octavian Udrea	<udrea@umiacs.umd.edu>
 *  Cristian Lumezanu	<lume@cs.umd.edu>
 *  Jeffrey S. Foster	<jfoster@cs.umd.edu>
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. The names of the contributors may not be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *)

(* rmcfind.ml : a simple set of functions: given a list of starting positions (integers) and a out and in pattern, it computes the list of instructions that can be used
   to run dataflow analysis on for a certain rule *)

open Pretty
open Cil
open Types

(* This function takes a list of Cil.globals and attempts to find the main function *)
let rec locateMain_Globals (g: Cil.global list) : Cil.fundec = 
    match g with
    [] -> failwith "Cannot locate main function !!"
    | h::r -> (match h with
	GFun(func,_) -> if func.svar.vname = "main" then
			    func
			else locateMain_Globals r
	| _ -> locateMain_Globals r	
    )
    ;;

(* This function locates the main function definition given a file *)
let locateMain (f: Cil.file) : Cil.fundec = 
    locateMain_Globals f.globals
    ;;
    
(* Given a list of statements, find the statement with a specified ID *)
let rec locateStatement (s: Cil.stmt list) (id: int) : Cil.stmt = 
    match s with
    [] -> Printf.printf "Statement with ID %d not found...\n" id; failwith "Statement not found!"
    | h::r -> if h.sid = id then h else locateStatement r id
    ;;

(* Given a function definition, finds and returns the first statement in the body of the function - i.e. the one with ID 0 *)
let locateFirstStatement (f: Cil.fundec) : Cil.stmt = 
    locateStatement f.sallstmts 0
    ;;

(* Verifies if two statements are equal by checking their ID *)
let equalsStmts (s1: Cil.stmt) (s2: Cil.stmt) = 
    if s1.sid = s2.sid then true else false
    ;;

(* This function checks whether a statment s is in a statement list *)
let rec isStmtInList (s: Cil.stmt) (slist: Cil.stmt list) =
    match slist with
    [] -> false
    | h::r -> if equalsStmts h s then true else isStmtInList s r
    ;;

(* This function concatenates two lists of statements making sure there are no duplicates *)
let rec concatStmtList (l1: Cil.stmt list) (l2: Cil.stmt list) : (Cil.stmt list) =
    match l1 with
    [] -> l2
    | h::r -> if isStmtInList h l2 then concatStmtList r l2
	      else concatStmtList r (h::l2)
    ;;

(* This function checks to see whether an expression is actually a constant string equal to the one given *)
let isExpString (e: Cil.exp) (str: string) = 
    match e with
    Const(c) -> (match c with
		    CStr(s) -> if s = str then true else false
		    | _ -> false
		   )
    | _ -> false
    ;;

(* This function checks to see if an instruction is a function call to fname *)
let isInstrFuncCall (ins: Cil.instr) (fname: string) = 
    match ins with
    Call(_,Lval(Var(vi),NoOffset),_,_) -> if vi.vname = fname then true else false
    | _ -> false
    ;;

(* This function checks to see whether a function call to fname is contained within the list of instructions *)
let rec isFuncCallInInstrList (l: Cil.instr list) (fname: string) = 
    match l with
    [] -> false
    | h::r -> if isInstrFuncCall h fname then true
	      else isFuncCallInInstrList r fname
    ;;

(* Checks whether a simple statement contains a function call *)
let rec isFuncCallInSimpleStmt (s: Cil.stmt) (fname: string) = 
    match s.skind with
    Instr(l) -> isFuncCallInInstrList l fname
    | _ -> false
    ;;

(* Checks whether a statement is a return with an integer value *)
let isStmtRetInt (s: Cil.stmt) (x: int) =
    match s.skind with
     Return(y,_) -> begin
    		     match y with
    		     Some e -> begin
		    		    match e with
				      Const(CInt64(v,_,_)) -> if Int64.to_int(v) = x then true else false
				      | _ -> false
		    	       end
		     | None -> false 
		    end    
     | _ -> false
    ;;

(* Checks whether a simple statement (non-recursive) is conformant to a pattern *)
let isStmtInPattern (s: Cil.stmt) (p: pattern) = 
    match p with
    InPat(x,_) -> isFuncCallInSimpleStmt s x.name
    | OutPat(x,_) -> isFuncCallInSimpleStmt s x.name
    | SuccPatF(x,_) -> isFuncCallInSimpleStmt s x.name
    | FailPatF(x,_) -> isFuncCallInSimpleStmt s x.name
    | SuccPatR(x) -> isStmtRetInt s x
    | FailPatR(x) -> isStmtRetInt s x
    ;;

(* Checks to see whether a statement corresponds to any pattern from the list *)
let rec isStmtInPatternList (s: Cil.stmt) (p: pattern list) = 
    match p with
    [] -> false
    | h::r -> if isStmtInPattern s h then true
	      else isStmtInPatternList s r
    ;;
    
(* Checks to see whether all statements are in the pattern list *)
let rec allStmtInPatternList (slist: Cil.stmt list) (p: pattern list) = 
    match slist with
    [] -> true
    | h::r -> if isStmtInPatternList h p then allStmtInPatternList r p
	      else false
    ;;
    
(* Given a set of statements, this function converts it to a set of simple (non-recursive) statements *)
let rec toSimpleStmtList (slist: Cil.stmt list) : (Cil.stmt list) =
    match slist with
    [] -> []
    | h::r -> (concatStmtList (toSimpleStmt h) (toSimpleStmtList r))
and
(* Given a statement, this function converts it recursively to a set of simple (non-recursive) statements *)
toSimpleStmt (s: Cil.stmt) : (Cil.stmt list) = 
    match s.skind with
    If(_,b1,b2,_) -> concatStmtList (toSimpleStmtList b1.bstmts) (toSimpleStmtList b2.bstmts)
    | Switch(_,b,l,_) -> toSimpleStmtList b.bstmts
    | Loop(b,_,_,_) -> toSimpleStmtList b.bstmts
    | Block(b) -> toSimpleStmtList b.bstmts
    | TryFinally(b1,b2,_) -> concatStmtList (toSimpleStmtList b1.bstmts) (toSimpleStmtList b2.bstmts)
    | TryExcept(b1,(_,_),b2,_) -> (concatStmtList (toSimpleStmtList b1.bstmts) (toSimpleStmtList b2.bstmts))
    | _ -> s::[]
    ;;

(* This function is the recursive version of the one below *)
let rec rmcFind_StmtList_rec (immutablepos: Cil.stmt list) (pos: Cil.stmt list) (pat: pattern list) (crStmts: Cil.stmt list) (x: int) : (Cil.stmt list)= 
    match pos with
    [] -> crStmts
    | h::r -> if isStmtInPatternList h pat then
		if isStmtInList h immutablepos then
		    if x > 0 then
			rmcFind_StmtList_rec immutablepos r pat (h::crStmts) x
		    else
			rmcFind_StmtList_rec immutablepos (concatStmtList r (toSimpleStmtList h.succs)) pat crStmts (x + 1)
		else rmcFind_StmtList_rec immutablepos r pat (h::crStmts) x
	      else rmcFind_StmtList_rec immutablepos (concatStmtList r (toSimpleStmtList h.succs)) pat crStmts (x + 1)
    ;;

(* Given a starting list of statements in the CFG, an in pattern and an out pattern, find all statements
   that we should run dataflow on starting from the given positions until an end pattern is reached. 
   NOTE: Whenever we reach the input pattern again, we will stop running through the statements (the same goes for success and fail patterns
   NOTE2: We assume that we are the very bottom level of the statement tree chain: that is our statements are not recursive
   -i.e. do not contain other statements.
*)
let rmcFind_StmtList (pos: Cil.stmt list) (pat: pattern list): (Cil.stmt list) =
    rmcFind_StmtList_rec pos pos pat [] 0
    ;;