(*
 *
 * Copyright (c) 2006-2007, 
 *  Octavian Udrea	<udrea@umiacs.umd.edu>
 *  Cristian Lumezanu	<lume@cs.umd.edu>
 *  Jeffrey S. Foster	<jfoster@cs.umd.edu>
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. The names of the contributors may not be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *)

(* rmcutils.ml: Contains general purpose utility functions - for instance obtaining a fact from an instruction or a set of facts from a simple instruction list, etc. *)

open Types
open Eunify
open Cil

(* Function checks whether an allVar appears in a list of allVars *)
let rec isInListAllVar (v: allVar) (vlist: allVar list) = 
    match vlist with
    [] -> false
    | h::r -> if equalsAllVar h v then
		true
	      else isInListAllVar v r
    ;;

(* Function concatenates two lists of allVars eliminating duplicates *)
let rec concatAllVarList (v1: allVar list) (v2: allVar list) = 
    match v1 with
    [] -> v2
    | h::r -> if isInListAllVar h v2 then
		concatAllVarList r v2
	      else concatAllVarList r (h::v2)
    ;;

(* Function checks whether an element is in a list *)
let rec isInListString (e: string) (l: string list) = 
    match l with
    [] -> false
    | h::r -> if h = e then true
	      else isInListString e r
    ;;

(* Concatenates two string lists *)
let rec concatStringList (v1: string list) (v2: string list) = 
    match v1 with
    [] -> v2
    | h::r -> if isInListString h v2 then
		concatStringList r v2
	      else concatStringList r (h::v2)
    ;;


(* This verifies that an expression is a constant int or not *)
let isExpInt (e: Cil.exp) = 
    match e with
    Const(c) -> 
	begin
	    match c with 
	    CInt64(_,_,_) -> true
	    | _ -> false
	end
    | _ -> false
    ;;

(* This function checks whether a lval is acceptable under the rules stated below *)
let checkLval (e: Cil.lval) (inname: string list) (outname: string list) = 
    match e with
    (lh,off) -> 
	let isOk = 
	    (match lh with
		Var(vi) -> (isInListString vi.vname inname) || (isInListString vi.vname outname)
		| _ -> false)
	in
	begin
	    match off with
	    NoOffset -> 
		begin
		    match lh with
			Var(vi) -> true
			| _ -> false
		end
	    | Field(_,off2) ->
		begin
		    match off2 with
		    NoOffset -> isOk
		    | Field(_,_) -> false
		    | Index(e,off3) -> 
			begin
			    match off3 with
			    NoOffset -> isOk && (isExpInt e)
			    | _ -> false
			end
		end
	    | _ -> false
	end
    ;;

(* This function checks whether a binary operator is acceptable to our rules *)
let checkBinOp (op: Cil.binop) = 
    match op with
    PlusA -> true
    | MinusA -> true
    | Mult -> true
    | Div -> true
    | _ -> false
    ;;

(* This function checks whether an exp can be transformed into a mixedExpr based on the rules below *)
let rec checkExpToMixedExpr (e: Cil.exp) (inname: string list) (outname: string list) = 
    match e with
    Const(c)->
	begin
	    match c with
		CInt64(_,_,_) -> true
		| _ -> false
	end
    | Lval(lv) -> checkLval lv inname outname
    | BinOp(op,e1,e2,_) -> (checkBinOp op) && (checkExpToMixedExpr e1 inname outname) && (checkExpToMixedExpr e2 inname outname)
    | _ -> false
    ;;

(* This function obtains an int assuming the expression is an int or -1 otherwise *)
let expToInt (e: Cil.exp) : int = 
    match e with
    Const(c) -> 
	begin
	    match c with 
	    CInt64(v,_,_) -> Int64.to_int(v)
	    | _ -> -1
	end
    | _ -> -1
    ;;

(* This function changes a valid (according to our rules) lval into a mixedExpr *)
let lvalToMixedExpr (lv: Cil.lval) (inname: string list) (outname: string list): mixedExpr = 
    match lv with
    (lh,off) -> 
	begin
	    match lh with
	    Var(vi) -> 
		begin
		    match off with
		    NoOffset -> MEVar({name = vi.vname})
		    | Field(fi,off2) -> 
			begin 
			    match off2 with 
			    NoOffset -> if isInListString vi.vname inname then
					    MEInVar(InVar({name = vi.vname}))
					else
					    MEOutVar(OutVar({name = vi.vname}))
			    | Index(e,_) -> if isInListString vi.vname inname then
						MEInVar(ArrayInVar({name = vi.vname},(expToInt e)))
					    else 
						MEOutVar(ArrayOutVar({name = vi.vname},(expToInt e)))
			    | _ -> failwith "Did not expect another field indirection in lvalToMixedExpr...\n"
			end
		    | _ -> failwith "Wrong lval in lvalToMixedExpr (2)...\n"
		end
	    | _ -> failwith "Wrong lval in lvalToMixedExpr (1)...\n"
	end
    ;;

(* This function changes a valid (according to our rules) lval into an allVar *)
let lvalToAllVar (lv: Cil.lval) (inname: string list) (outname: string list): allVar = 
    match lv with
    (lh,off) -> 
	begin
	    match lh with
	    Var(vi) -> 
		begin
		    match off with
		    NoOffset -> AVar({name = vi.vname})
		    | Field(fi,off) -> 
			begin 
			    match off with
			    NoOffset -> if isInListString vi.vname inname then
					    AInVar(InVar({name = vi.vname}))
					else
					    AOutVar(OutVar({name = vi.vname}))
			    | Index(e,_) -> if isInListString vi.vname inname then
						AInVar(ArrayInVar({name = vi.vname},(expToInt e)))
					    else 
						AOutVar(ArrayOutVar({name = vi.vname},(expToInt e)))
			    | _ -> failwith "Did not expect another field indirection in lvalToAllVar...\n"
			end
		    | _ -> failwith "Wrong lval in lvalToAllVar (2)...\n"
		end
	    | _ -> failwith "Wrong lval in lvalToAllVar (1)...\n"
	end
    ;;
    
(* This function checks the rhs of an assignment (given as an exp) for whether we can use it as a fact. The rules are:
    - it only contains simple variables, structure fields that refer to either inname and outname or elements of the vectors of these or ints
    The function returns a mixedExpr if everything works out fine 
*)
let rec expToMixedExpr (e: Cil.exp) (inname: string list) (outname: string list) : mixedExpr = 
    match e with
    Const(c) -> 
	begin
	    match c with
		CInt64(v,_,_) -> MEInt(Int64.to_int(v))
		| _ -> failwith "Wrong constant type in expToMixedExpr...\n"
	end
    | Lval((lh,off)) -> lvalToMixedExpr (lh,off) inname outname
    | BinOp (op, e1, e2,_) ->
	let ee1 = expToMixedExpr e1 inname outname in
	    let ee2 = expToMixedExpr e2 inname outname in
		begin
		    match op with
		    PlusA -> MixedExpr(PlusOp,ee1,ee2)
		    | MinusA -> MixedExpr(MinusOp,ee1,ee2)
		    | Mult -> MixedExpr(TimesOp,ee1,ee2)
		    | Div -> MixedExpr(DivOp,ee1,ee2)
		    | _ -> failwith "Wrong operator in expToMixedExpr...\n"
		end
    | _ -> failwith "Wrong expression type in expToMixedExpr...\n"
    ;;

(* This function obtains a fact from an instruction if any *)
let getFact (i: Cil.instr) (inname: string list) (outname: string list) = 
    match i with
	Set(lv,e,_) -> if checkLval lv inname outname then
			if checkExpToMixedExpr e inname outname then
			    Fact(EqOp, (lvalToAllVar lv inname outname), (expToMixedExpr e inname outname))
			else Nofact
		   else Nofact
	| _ -> Nofact
    ;;

(* This function extracts a fact from a global (if the global contains an initializer) *)
let getFactFromGlobal (g: Cil.global) : fact = 
    match g with
    GVar(vi,inf,_) -> 
	(
	    match inf.init with
	    Some iinf -> 
		(match iinf with
		SingleInit(e) -> if checkExpToMixedExpr e [] [] then
		    		     Fact(EqOp,AVar({name = vi.vname}),(expToMixedExpr e [] []))
				 else Nofact
		| _ -> Nofact
		)
	    | None -> Nofact
	)
    | _ -> Nofact
    ;;

(* This function obtains all facts from a global list *)
let rec getFactFromGlobalList (g: Cil.global list): factList = 
    match g with
    [] -> []
    | h::r -> fact_add (getFactFromGlobalList r) (getFactFromGlobal h)
    ;;

(* Get the element count of a list *)
let rec getElementCount l = 
    match l with
    [] -> 0
    | h::r -> 1 + (getElementCount r)
    ;;

(* Given a list, find the x'th element in the list *)
let rec getListElement l x = 
    match l with
    [] -> failwith "Too few elements in getListElement...\n"
    | h::r -> if x = 0 then h else getListElement r (x - 1)
    ;;

(* Checks if an exp is in fact a simple variable *)
let checkExpToVar (e: Cil.exp) = 
    match e with 
    Lval((Var(_),NoOffset)) -> true
    | _ -> false
    ;;

(* Converts an exp is to a simple variable *)
let expToVar (e: Cil.exp) : string = 
    match e with 
    Lval((Var(vi),NoOffset)) -> vi.vname
    | _ -> ""
    ;;
    
(* Given an instruction, it does what the comments below indicate *)
let getFuncArgNameInstr (i: Cil.instr) (fname: string) (argc: int): (string list) = 
    match i with
    Call(_,Lval((Var vi),NoOffset),el,_) -> if vi.vname = fname then
						if (getElementCount el) >= argc then
						    let elem = getListElement el argc in
							let str = expToVar elem in
							    if str = "" then []
							    else str::[] 
						else []
					   else []
    | _ -> []
    ;;

(* Given a list of instructions, it does what you can see in the comments below *)
let rec getFuncArgNameInstrList (ilist: Cil.instr list) (fname: string) (argc: int) : (string list) = 
    match ilist with
    [] -> []
    | h::r -> concatStringList (getFuncArgNameInstr h fname argc) (getFuncArgNameInstrList r fname argc)
    ;;

(* Given a simple statement, a function name and an argument number, it returns the list of allVars *)
let getFuncArgNameStmt (s: Cil.stmt) (fname: string) (argc: int) : (string list) = 
    match s.skind with
    Instr(il) -> getFuncArgNameInstrList il fname argc
    | _ -> []
    ;;

(* Given a list of simple statements, we search for an instruction that is a function call to fname and try to obtain the x's argument name (if it is a variable) *)
let rec getFuncArgName (slist: Cil.stmt list) (fname: string) (argc: int) : (string list) = 
    match slist with
    [] -> []
    | h::r -> concatStringList (getFuncArgNameStmt h fname argc) (getFuncArgName r fname argc) 
    ;;

(* Given a set of simple statements, this method finds and returns the name of variables (if any) matching a pattern. It only works for input or output patterns *)
let getPatternNames (slist: Cil.stmt list) (p: pattern) : (string list) = 
    let fname = 
    (match p with
     InPat(v,_) -> v.name
     | OutPat(v,_) -> v.name
     | _ -> ""
    ) in 
	if fname = "" then []
	else 
	    let argc = 
	    (match p with
		InPat(_,x) -> x
		| OutPat(_,x) -> x
		| _ -> 0
	    ) in
		if argc = 0 then [] 
		else getFuncArgName slist fname argc
    ;;