(*
 *
 * Copyright (c) 2006-2007, 
 *  Octavian Udrea	<udrea@umiacs.umd.edu>
 *  Cristian Lumezanu	<lume@cs.umd.edu>
 *  Jeffrey S. Foster	<jfoster@cs.umd.edu>
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. The names of the contributors may not be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *)

open Types

(************************************************* GENERAL DEBUGGING FUNCTIONS ***********************************************)

(* prints a string list *)
let rec print_StringList (s: string list) = 
    match s with
    [] -> Printf.printf "\n"
    | h::r -> Printf.printf "%s " h;
	      print_StringList r
    ;;

(* Prints out a CFG, each line containing info about a node *)
let rec printCFG (cfg: ruleCFG) = 
    match cfg with 
    [] -> Printf.printf "All done!\n";
    | h::r -> 
	Printf.printf "Node: %s; predecessors: " h.ruleName;
	print_StringList h.pred;
	printCFG r
    ;;
    
(* Prints out a variable list *)
let rec print_VarList (l : var list) =
    match l with
    [] -> print_string "\n"
    | h::r -> print_string h.name;
	      print_string " ";
	      print_VarList r
    ;;
    
(************************************ LIST AND SET UTILITY METHODS *********************************)

(* Check if a list contains an element *)
let rec list_contains l e = 
	match l with
	[] -> false
	| h::r -> if h = e then true else list_contains r e
	;;

(* Append an element to a list if it is not already there *)
let set_append l e = 
	if list_contains l e then
		l
	else e::l
	;;

(* concatenates two lists *)
let rec set_concat l1 l2 = 
	match l1 with
	[] -> l2
	| h::r -> set_concat r (set_append l2 h)
	;;
	
(* Determines the number of elements in a list *)
let rec list_count l = 
    match l with
    [] -> 0
    | h::r -> 1 + (list_count r)
    ;;

(* determines if the intersection of two sets is empty or not *)
let rec intersection_empty l1 l2 = 
    match l1 with
    [] -> true
    | h::r -> if list_contains l2 h then
		false
	      else intersection_empty r l2
    ;;

(************************************ RULE UTILITY METHODS *****************************************)

(* Searches a rule list for a rule with the specified name *)
let rec getRuleByName (rules: rule list) (rulename: string) =
    match rules with
    [] -> failwith "Rule not found in rule list...\n"
    | h::r -> if h.rname.name = rulename then
                h
              else getRuleByName r rulename
    ;;

(* Searches an inExpression for used rule variables *)
let rec getUsedRuleVars_inExpr (e: inExpr) = 
    match e with
    EInVar(_) -> []
    | EInt(_) -> []
    | ERuleVar(v) -> (match v with
		       RuleVar(x) -> [x]
		       | ArrayRuleVar(x,_) -> [x])
    | InExpr(_,e1,e2) -> set_concat (getUsedRuleVars_inExpr e1) (getUsedRuleVars_inExpr e2)
    ;;

(* Searches a premise for used rule variables *)
let getUsedRuleVars_Premise (p: premise) = 
    match p with
    Premise(_,_,e) -> getUsedRuleVars_inExpr e
    | RVarPremiseStr(v,_) -> (match v with
			     RuleVar(x) -> [x]
			     | ArrayRuleVar(x,_) -> [x])
    | RVarPremiseInt(v,_) -> (match v with
			     RuleVar(x) -> [x]
			     | ArrayRuleVar(x,_) -> [x])
    ;;

(* Searches a premise list and finds the rule variables that are used, but not set *)
let rec getUsedRuleVars_PremiseList (p: premise list) = 
    match p with 
    [] -> []
    | h::r -> set_concat (getUsedRuleVars_Premise h) (getUsedRuleVars_PremiseList r)
    ;;

(* Searches a conclusion for used rule variables *)
let getUsedRuleVars_Conclusion(c : conclusion) = 
    match c with
    Conclusion(_,_,e) -> getUsedRuleVars_inExpr e
    | Set(_,_) -> []
    | NoOut -> []
    | Success -> []
    | Fail -> []
    ;;

(* Searches a conclusion list and finds the rule variables that are used, but not set *)
let rec getUsedRuleVars_ConclusionList (p: conclusion list) = 
    match p with 
    [] -> []
    | h::r -> set_concat (getUsedRuleVars_Conclusion h) (getUsedRuleVars_ConclusionList r)
    ;;
    
(* Given a rule, it gives a set of all rule variables that appear in the
   premises or conclusions except the ones that appear in set statements *)
let getUsedRuleVars (r : rule) =
    set_concat (getUsedRuleVars_PremiseList r.premises) (getUsedRuleVars_ConclusionList r.conclusions)
    ;;

(* Determine rule variables set in a conclusion *)
let getSetRuleVars_Conclusion (c: conclusion) = 
    match c with
    Conclusion(_,_,_) -> []
    | Set(v,_) -> (match v with
			     RuleVar(x) -> [x]
			     | ArrayRuleVar(x,_) -> [x])
    | NoOut -> []
    | Success -> []
    | Fail -> []
    ;;

(* Determines rule variables set in a conclusion list *)
let rec getSetRuleVars_ConclusionList (c: conclusion list) = 
    match c with 
    [] -> []
    | h::r -> set_concat (getSetRuleVars_Conclusion h) (getSetRuleVars_ConclusionList r)
    ;;
    
(* Determines rule variables set by a rule *)
let getSetRuleVars (r: rule) =
    getSetRuleVars_ConclusionList r.conclusions
    ;;

(* obtains and returns the predecessors of a certain rule from a rule list in the rule CFG *)
let rec getPredecessors_RuleList (rules:rule list) (r: rule) = 
    match rules with
    [] -> []
    | h::s -> 
	if intersection_empty (getSetRuleVars h) (getUsedRuleVars r) then
		getPredecessors_RuleList s r
	    else
		(h.rname.name)::(getPredecessors_RuleList s r)
    ;;

(* creates and returns a CFG Node for a certain rule *)
let createCFGNode (rules: rule list) (r: rule) = 
    { ruleName = r.rname.name;
      pred = (getPredecessors_RuleList rules r)
    }
;;

(* creates a CFG from a list of rules *)
let rec createCFG_RuleList (rules: rule list) (crules: rule list) = 
    match crules with
    [] -> []
    | h::r -> (createCFGNode rules h)::(createCFG_RuleList rules r)
    ;;

(* creates and returns the CFG for the rules. The CFG is a list of nodes,
   each node containing the name of the rule, its successors and predecessors
   (names) *)
let createRuleCFG (ruleprog: program) =
	createCFG_RuleList ruleprog.rules ruleprog.rules
    ;;

(* searches for a rule's predecessors given its name *)
let rec getCFGPredecessors (cfg: ruleCFG) (rname: string) : (string list) = 
    match cfg with
    [] -> []
    | h::r -> if h.ruleName = rname then h.pred 
	      else getCFGPredecessors r rname
    ;;